/* linord.c (linear ordering problem solver) */

/* Author: Andrew Makhorin <mao@gnu.org>, October 2009 */

#include <stdio.h>
#include <stdlib.h>

#include <glpk.h>
#define xassert glp_assert
#define xprintf glp_printf

/***********************************************************************
*  LINEAR ORDERING PROBLEM
*
*  Let G = (V,E) denote the complete digraph, where V is the set of
*  nodes and E is the set of arcs. A tournament T in E consists of a
*  subset of arcs containing for every pair of nodes i and j either arc
*  (i->j) or arc (j->i), but not both. T is an acyclic tournament if it
*  contains no directed cycles. Obviously, an acyclic tournament induces
*  an ordering <i1, i2, ..., in> of the nodes (and vice versa), where
*  n = |V|. Node i1 is the one with no entering arcs, i2 has exactly one
*  entering arc, etc., and in is the node with no outgoing arc. Given
*  arc weights w[i,j] for every pair i, j in V, the Linear Ordering
*  Problem (LOP) consists of finding an acyclic tournament T in E such
*  that the sum of arcs in T is maximal, or in other words, of finding
*  an ordering of the nodes such that the sum of the weights of the arcs
*  compatible with this ordering is maximal.
*
*  Given a nxn-matrix C = (c[i,j]) the triangulation problem is to
*  determine a symmetric permutation of the rows and columns of C such
*  that the sum of subdiagonal entries is as small as possible. Note
*  that it does not matter if diagonal entries are taken into account or
*  not. Obviously, by setting arc weights w[i,j] = c[i,j] for the
*  complete digraph G, the triangulation problem for C can be solved as
*  linear ordering problem for G. Conversely, a linear ordering problem
*  for G can be transformed to a triangulation problem for an nxn-matrix
*  C by setting c[i,j] = w[i,j] and the diagonal entries c[i,i] = 0 (or
*  to arbitrary values).
*
*  The LOP can be formulated as binary integer programming problem.
*  We use binary variables x[i,j] for (i,j) in A, stating whether arc
*  (i->j) is present in the tournament or not. Taking into account that
*  a tournament is acyclic iff it contains no dicycles of length 3, it
*  is easy to see that the LOP can be formulated as follows:
*
*     Maximize
*
*        sum    w[i,j] x[i,j]                                        (1)
*     (i,j) in A
*
*     Subject to
*
*        x[i,j] + x[j,i] = 1, for all i,j in V, i < j,               (2)
*
*        x[i,j] + x[j,k] + x[k,i] <= 2,                              (3)
*
*              for all i,j,k in V, i < j, i < k, j != k,
*
*        x[i,j] in {0, 1}, for all i,j in V.                         (4)
*
*  (From <http://heur.uv.es/optsicom/LOLIB/>.) */

#define N_MAX 100
/* maximal number of nodes */

int n;
/* number of nodes in the digraph given */

int w[1+N_MAX][1+N_MAX];
/* w[i,j] is weight of arc (i->j), 1 <= i,j <= n */

/***********************************************************************
*  read_data - read LOP instance data
*
*  This routine reads LOP instance data from a text file. The following
*  format is used:
*
*     comment line
*     n
*     w[1,1] w[1,2] ... w[1,n]
*     w[2,1] w[2,2] ... w[2,n]
*     . . . . .
*     w[n,1] w[n,2] ... w[n,n] */

void read_data(char *fname)
{     glp_data *data;
      const char *comment;
      int i, j;
      xprintf("Reading LOP instance data from `%s'...\n", fname);
      data = glp_sdf_open_file(fname);
      if (data == NULL) exit(EXIT_FAILURE);
      comment = glp_sdf_read_text(data);
      xprintf("%s\n", comment);
      n = glp_sdf_read_int(data);
      if (n < 1)
         glp_sdf_error(data, "invalid number of nodes\n");
      if (n > N_MAX)
         glp_sdf_error(data, "too many nodes\n");
      xprintf("Digraph has %d nodes\n", n);
      for (i = 1; i <= n; i++)
      {  for (j = 1; j <= n; j++)
            w[i][j] = glp_sdf_read_int(data);
      }
      xprintf("%d lines were read\n", glp_sdf_line(data));
      glp_sdf_close_file(data);
      return;
}

glp_prob *P;
/* problem object */

int x[1+N_MAX][1+N_MAX];
/* x[i][j] is the number of binary variable x[i,j], 1 <= i,j <= n,
   i != j, in the problem object, where x[i,j] = 1 means that node i
   precedes node j, i.e. arc (i->j) is included in the tournament */

/***********************************************************************
*  build_mip - build initial MIP formulation
*
*  This routine builds the initial MIP formulation, which includes the
*  binary variables x[i,j] and irreflexivity constraints (2) only. */

void build_mip(void)
{     int i, j, row, ind[1+2];
      double val[1+2];
      char name[50+1];
      P = glp_create_prob();
      glp_set_obj_dir(P, GLP_MAX);
      /* create binary variables */
      for (i = 1; i <= n; i++)
      {  for (j = 1; j <= n; j++)
         {  if (i == j)
            {  x[i][j] = 0;
               continue;
            }
            x[i][j] = glp_add_cols(P, 1);
            sprintf(name, "x[%d,%d]", i, j);
            glp_set_col_name(P, x[i][j], name);
            glp_set_col_kind(P, x[i][j], GLP_BV);
            /* objexctive coefficient */
            glp_set_obj_coef(P, x[i][j], w[i][j]);
         }
      }
      /* create irreflexivity constraints (2) */
      for (i = 1; i <= n; i++)
      {  for (j = i+1; j <= n; j++)
         {  row = glp_add_rows(P, 1);
            glp_set_row_bnds(P, row, GLP_FX, 1, 1);
            ind[1] = x[i][j], val[1] = 1;
            ind[2] = x[j][i], val[2] = 1;
            glp_set_mat_row(P, row, 2, ind, val);
         }
      }
      return;
}

int generate_rows(void)
{     /* generate violated transitivity constraints (3) and add them
         to the current subproblem; as suggested by Juenger et al., we
         consider only arc-disjoint violated constraints */
      int i, j, k, u[1+N_MAX][1+N_MAX], row, ind[1+3], cnt;
      double r, val[1+3];
      /* u[i,j] = 1, if arc (i->j) is covered by some constraint */
      for (i = 1; i <= n; i++)
      {  for (j = 1; j <= n; j++)
            u[i][j] = 0;
      }
      cnt = 0;
      for (i = 1; i <= n; i++)
      {  for (j = 1; j <= n; j++)
         {  for (k = 1; k <= n; k++)
            {  if (i == j) continue;
               if (i == k) continue;
               if (j == k) continue;
               if (u[i][j] || u[j][i]) continue;
               if (u[i][k] || u[k][i]) continue;
               if (u[j][k] || u[k][j]) continue;
               /* check if x[i,j] + x[j,k] + x[k,i] <= 2 */
               r = glp_get_col_prim(P, x[i][j]) +
                   glp_get_col_prim(P, x[j][k]) +
                   glp_get_col_prim(P, x[k][i]);
               /* should note that it is not necessary to add to the
                  current subproblem every violated constraint (3), for
                  which r > 2; if r < 3, we can stop adding violated
                  constraints, because for integer feasible solution
                  the value of r is integer, so r < 3 is equivalent to
                  r <= 2; on the other hand, adding violated
                  constraints leads to tightening the feasible region
                  of LP relaxation and, thus, may reduce the size of
                  the search tree */
               if (r <= 2.15) continue;
               /* generate violated transitivity constraint */
               row = glp_add_rows(P, 1);
               glp_set_row_bnds(P, row, GLP_UP, 0, 2);
               ind[1] = x[i][j], val[1] = 1;
               ind[2] = x[j][k], val[2] = 1;
               ind[3] = x[k][i], val[3] = 1;
               glp_set_mat_row(P, row, 3, ind, val);
               u[i][j] = u[j][i] = 1;
               u[i][k] = u[k][i] = 1;
               u[j][k] = u[k][j] = 1;
               cnt++;
            }
         }
      }
#if 0
      xprintf("%d violated constraints were generated\n", cnt);
#endif
      return cnt;
}

int inactive(glp_tree *T, int list[])
{     /* build the list of inactive transitivity constraints on the
         current level */
      glp_attr attr;
      int p = glp_ios_curr_node(T);
      int lev = glp_ios_node_level(T, p);
      int i, cnt = 0;
      for (i = glp_get_num_rows(P); i >= 1; i--)
      {  glp_ios_row_attr(T, i, &attr);
         if (attr.level < lev) break;
         if (attr.origin != GLP_RF_REG)
         {  if (glp_get_row_stat(P, i) == GLP_BS)
            {  cnt++;
               if (list != NULL) list[cnt] = i;
            }
         }
      }
      return cnt;
}

void remove_inactive(glp_tree *T)
{     /* remove inactive transitivity constraints */
      int cnt, *list;
      cnt = inactive(T, NULL);
      if (cnt > 0)
      {  list = glp_calloc(1+cnt, sizeof(int));
         inactive(T, list);
         glp_del_rows(P, cnt, list);
         glp_free(list);
#if 0
         xprintf("%d inactive constraints removed\n", cnt);
#endif
      }
      return;
}

void callback(glp_tree *T, void *info)
{     xassert(info == info);
      if (glp_ios_reason(T) == GLP_IROWGEN)
      {  remove_inactive(T);
         if (generate_rows() == 0)
            remove_inactive(T);
      }
      return;
}

int main(int argc, char *argv[])
{     if (argc != 2)
      {  xprintf("Usage: %s filename\n", argv[0]);
         return 0;
      }
      read_data(argv[1]);
      build_mip();
      glp_adv_basis(P, 0);
      glp_simplex(P, NULL);
      {  glp_iocp parm;
         glp_init_iocp(&parm);
         parm.cb_func = callback;
         parm.cb_info= NULL;
         glp_intopt(P, &parm);
      }
      glp_print_mip(P, "sol");
      return 0;
}

/* eof */
